package com.ibm.abbot.tester.swt;

import java.awt.Component;
import java.awt.Container;
import java.util.Hashtable;

import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.Control;

/**
 * A helper class with routines to make GUI testing easier.
 * 
 * @author Jeff Tulley
 */
public class GUITestCaseHelper
{
   ///////////////////////////////// SWING ///////////////////////////////
   /**
	* Goes through all of the child components of the container that is sent in
	* and puts the components into a hashtable with the name of the component 
	* as the key.
	* 
	* This is useful when you have retrieved a container of some sort from your
	* code being tested, and you want to verify certain conditions by looking 
	* up components in that container.
	* 
	* Note:  Using such a testing facility requires you to name all of your 
	* components, which is a good practice anyway.
	* 
	* @param inContainer
	*                A container to parse
	* @param recurse whether to recurse into child containers
	* 
	* @return A populated Hashtable.
	*/
   public static Hashtable parseComponents (Container inContainer,
											boolean recurse)
   {
	  return parseComponents(inContainer.getComponents(), recurse);
   }

   /**
	* Goes through all of the components sent in and puts them into a hashtable
	* with the name of the component as the key.
	* 
	* This is useful when you have retrieved a container of some sort from your
	* code being tested, and you want to verify certain conditions by looking 
	* up components in that container.
	* 
	* Note:  Using such a testing facility requires you to name all of your 
	* components, which is a good practice anyway.
	* 
	* @param inComponents
	*                An array of Components to parse
	* @param recurse whether to recurse into child containers
	* 
	* @return A populated Hashtable.
	*/
   public static Hashtable parseComponents (Component[] inComponents, 
											boolean recurse)
   {
	  Hashtable components = new Hashtable();
	  components.clear();
	  parseComponents(components, inComponents,recurse);
	  return components;
   }

   /**
	* Goes through all of the components sent in and puts them into the 
	* hashtable with the name of the component as the key.
	* This is useful when you have retrieved a container of some sort from
	* your code being tested, and you want to verify certain conditions by
	* looking up components in that container.
	* 
	* If you pass in recurse = true, then if some of the passed-in components
	* are actually containers themselves, their children will also be added
	* to the hash recursively.
	* 
	* Generally you can use on of the other parse routines easiest.
	* Note:  Using such a testing facility requires you to name all of your 
	* components, which is a good practice anyway.
	* 
	* @param inHash  Hashtable to be populated
	* @param inComponents
	*                An array of Components to parse
	* @param recurse whether to recurse into child containers
	* 
	* @return The return value only exists on this method to be type
	*         compatible with the other two parseComponents methods.
	*         Since the Hashtable is passed in, the return value isn't
	*         strictly necessary.
	*/
   public static Hashtable parseComponents (Hashtable inHash, 
											Component[] inComponents, 
											boolean recurse)
   {
	  for (int i = 0; i < inComponents.length; i++)
	  {
		 Component curComp = inComponents[i];
		 String name = curComp.getName();
		 if (curComp instanceof Container && recurse == true)
		 {
			parseComponents(inHash, ((Container)curComp).getComponents(), 
							recurse);
		 }
		 if (name != null)
		 {
			inHash.put(name, curComp);
		 } 
	  }
	  // only doing a return value here so we can have one method name with the
	  // same return type across the board
	  return inHash;
   }


   ///////////////////////////////// SWT ///////////////////////////////
   /**
	* Goes through all of the child components of the container that is sent in
	* and puts the components into a hashtable with the name of the component 
	* as the key.
	* 
	* This is useful when you have retrieved a container of some sort from your
	* code being tested, and you want to verify certain conditions by looking 
	* up components in that container.
	* 
	* Note:  Using such a testing facility requires you to name all of your 
	* components.  In SWT this is done with 'setData("name", "SomeName");'
	* 
	* @param inContainer
	*                A container to parse
	* @param recurse whether to recurse into child containers
	* 
	* @return A populated Hashtable.
	*/
   public static Hashtable parseComponents (Composite inContainer,
											boolean recurse)
   {
	  return parseComponents(inContainer.getChildren(), recurse);
   }

   /**
	* Goes through all of the components sent in and puts them into a hashtable
	* with the name of the component as the key.
	* 
	* This is useful when you have retrieved a container of some sort from your
	* code being tested, and you want to verify certain conditions by looking 
	* up components in that container.
	* 
	* Note:  Using such a testing facility requires you to name all of your 
	* components.  In SWT this is done with 'setData("name", "SomeName");'
	* 
	* @param inComponents
	*                An array of Components to parse
	* @param recurse whether to recurse into child containers
	* 
	* @return A populated Hashtable.
	*/
   public static Hashtable parseComponents (Control[] inComponents, 
											boolean recurse)
   {
	  Hashtable components = new Hashtable();
	  components.clear();
	  parseComponents(components, inComponents,recurse);
	  return components;
   }

   /**
	* Goes through all of the components sent in and puts them into the 
	* hashtable with the name of the component as the key.
	* This is useful when you have retrieved a container of some sort from
	* your code being tested, and you want to verify certain conditions by
	* looking up components in that container.
	* 
	* If you pass in recurse = true, then if some of the passed-in components
	* are actually containers themselves, their children will also be added
	* to the hash recursively.
	* 
	* Generally you can use on of the other parse routines easiest.
	* Note:  Using such a testing facility requires you to name all of your 
	* components.  In SWT this is done with 'setData("name", "SomeName");'
	* 
	* @param inHash  Hashtable to be populated
	* @param inComponents
	*                An array of Components to parse
	* @param recurse whether to recurse into child containers
	* 
	* @return The return value only exists on this method to be type
	*         compatible with the other two parseComponents methods.
	*         Since the Hashtable is passed in, the return value isn't
	*         strictly necessary.
	*/
   public static Hashtable parseComponents (Hashtable inHash, 
											Control[] inComponents, 
											boolean recurse)
   {
	  for (int i = 0; i < inComponents.length; i++)
	  {
		 Control curComp = inComponents[i];
		 String name = (String)curComp.getData("name");
		 if (curComp instanceof Composite && recurse == true)
		 {
			parseComponents(inHash, ((Composite)curComp).getChildren(), 
							recurse);
		 }
		 if (name != null)
		 {
			inHash.put(name, curComp);
		 } 
	  }
	  // only doing a return value here so we can have one method name with the
	  // same return type across the board
	  return inHash;
   }
}
