
package com.ibm.abbot.tester.swt;

import org.eclipse.swt.widgets.*;
import org.eclipse.swt.graphics.*;
import org.eclipse.swt.internal.win32.*;
import org.eclipse.swt.*;

/**
 * Provides widget-specific actions for testing a scrollBar.
 * 
 * NOTE: All methods are platform-dependent, and as of now, have 
 * only been tested on Windows 2000.  They are only advertised to work
 * on Windows 2000 with SWT version 2.1.1. 
 */ 
public class ScrollBarTester extends WidgetTester {
	/* Widget properties that are obtainable by member getter functions */
	protected Scrollable parent;
	protected int minimum, maximum, increment, pageIncrement, thumb, selection;
	protected boolean visible, enabled;
	protected Point size;

	/**
	 * Sets the above properties to their current values for the given widget. 
	 * NOTE: This should be called in a block of code synchronized on this
	 * tester.
	 */
	protected synchronized void getProperties(final ScrollBar bar){
		super.getProperties(bar);
		Robot.syncExec(bar.getDisplay(),this,new Runnable(){
			public void run(){
				parent = bar.getParent();
				minimum = bar.getMinimum();
				maximum = bar.getMaximum();
				increment = bar.getIncrement();
				pageIncrement = bar.getPageIncrement();
				size = bar.getSize();
				thumb = bar.getThumb();
				selection = bar.getSelection();
				visible = bar.getVisible();
				enabled = bar.getEnabled();
			}
		});
	}

	/**
	 * These getter methods return a particular property of the given widget.
	 * @see the corresponding member function in class Widget   
	 */ 
	/* Begin getters */
	public synchronized boolean getEnabled(ScrollBar bar) {
		getProperties(bar);
		return enabled;
	}

	public synchronized int getIncrement(ScrollBar bar) {
		getProperties(bar);
		getProperties(bar);
		return increment;
	}

	public synchronized int getMaximum(ScrollBar bar) {
		return maximum;
	}

	public synchronized int getMinimum(ScrollBar bar) {
		getProperties(bar);
		return minimum;
	}

	public synchronized int getNewIncrement(ScrollBar bar) {
		getProperties(bar);
		return newIncrement;
	}

	public synchronized int getNewSelection(ScrollBar bar) {
		getProperties(bar);
		return newSelection;
	}

	public synchronized int getPageIncrement(ScrollBar bar) {
		getProperties(bar);
		return pageIncrement;
	}

	public synchronized Scrollable getParent(ScrollBar bar) {
		getProperties(bar);
		return parent;
	}

	public synchronized int getSelection(ScrollBar bar) {
		getProperties(bar);
		return selection;
	}

	public synchronized Point getSize(ScrollBar bar) {
		getProperties(bar);
		return size;
	}

	public synchronized int getStyle(ScrollBar bar) {
		getProperties(bar);
		return style;
	}

	public synchronized int getThumb(ScrollBar bar) {
		getProperties(bar);
		return thumb;
	}

	public synchronized boolean getVisible(ScrollBar bar) {
		getProperties(bar);
		return visible;
	}
	/* End getters */

	int style, newIncrement, newSelection;

	// platform-dependent fields based on the rendering of the scrollbar
	public final static int BUTTON_SIZE = 17;
	public final static int THUMB_OFFSET = 2;

	/** 
	 * Scroll the given amount, where amount==the number of 
	 * times the arrow is pressed.  Positive amounts scroll down 
	 * for vertical scrollbars and right for horizontal ones.
	 */
	public void actionScroll(final ScrollBar bar, final int amount){
		actionFocus(bar);
		final Rectangle bounds = getBounds(bar);
		System.out.println("BOUNDS: "+bounds);
		if(amount<0){
			mouseMove(bounds.x+BUTTON_SIZE/2,bounds.y+BUTTON_SIZE/2);
			for(int i=0; i<amount;i++){
				mousePress(SWT.BUTTON1);
				mouseRelease(SWT.BUTTON1);	
			}				
		}
		else if(amount>0){
			mouseMove(	bounds.x+bounds.width - BUTTON_SIZE/2,
						bounds.y+bounds.height - BUTTON_SIZE/2);
			for(int i=0; i<amount;i++){
				mousePress(SWT.BUTTON1);
				mouseRelease(SWT.BUTTON1);	
			}		
		}	
		actionWaitForIdle(bar.getDisplay());
	}

	/**
	 * Page the given amount, where amount==# of times that the
	 * bar is clicked in the page area
	 **/
	public synchronized void actionPageScroll(final ScrollBar bar, int amount){
		actionFocus(bar);
		final Rectangle bounds = getBounds(bar);
		//System.out.println("BOUNDS: "+bounds);
		Robot.syncExec(bar.getDisplay(),this,new Runnable(){
			public void run(){
				style = bar.getStyle();
			}			
		});
		
		if(amount<0){
			if((style&SWT.HORIZONTAL)==SWT.HORIZONTAL)
				mouseMove(bounds.x+BUTTON_SIZE,bounds.y+BUTTON_SIZE/2);
			else
				mouseMove(bounds.x+BUTTON_SIZE/2,bounds.y+BUTTON_SIZE);
			for(int i=0; i<amount;i++){
				mousePress(SWT.BUTTON1);
				mouseRelease(SWT.BUTTON1);	
			}								
		}
		else if(amount > 0){
			if((style&SWT.HORIZONTAL)==SWT.HORIZONTAL)
				mouseMove(	bounds.x+bounds.width - BUTTON_SIZE,
							bounds.y+bounds.height - BUTTON_SIZE/2);
			else
				mouseMove(	bounds.x+bounds.width - BUTTON_SIZE/2,
							bounds.y+bounds.height - BUTTON_SIZE);				
			for(int i=0; i<amount;i++){
				mousePress(SWT.BUTTON1);
				mouseRelease(SWT.BUTTON1);	
			}					
		}
		actionWaitForIdle(bar.getDisplay());
	}

	/** 
	 * Sets the selection to the given value, or as close as possible, by dragging
	 * the slider.  The smaller the ratio of bar.getThumb()/(bar.getMaximum()-bar.getMinimum()),
	 * the less accurate this method is.
	 **/
	// TODO FIXME method loses the lock now between calls to syncExec, so 
	// we need to do everything in one big syncExec block
	public synchronized void actionScrollSetSelection(final ScrollBar bar, final int val){
		actionFocus(bar);
		final Rectangle bounds = getBounds(bar);
		// get info about current selection
		Robot.syncExec(bar.getDisplay(),this,new Runnable(){
			public void run(){
				style = bar.getStyle();
				selection = bar.getSelection();
				increment = bar.getIncrement();
				minimum = bar.getMinimum();
				maximum = bar.getMaximum();
				thumb = bar.getThumb();		
			}
		});
		actionWaitForIdle(bar.getDisplay());
		int setTo = val;
		if(setTo<minimum)setTo = minimum;
		else if(setTo>maximum-thumb)setTo = maximum-thumb;
		
		if((style&SWT.HORIZONTAL)==SWT.HORIZONTAL){		
			double thumbEdge = (double)((double)selection/(double)(maximum-minimum));
			double thumbWidth = (double)((double)thumb/(double)(maximum-minimum));
			double setEdge = (double)((double)setTo/(double)(maximum-minimum));
			double delta = (double)((double)increment/(double)(maximum-minimum));
			
			int stripWidth = bounds.width - 2*BUTTON_SIZE;
			int thumbLocPixels = (int)(thumbEdge*stripWidth);
			int thumbWidthPixels = (int)(thumbWidth*stripWidth);
			int setEdgePixels = (int)(setEdge*stripWidth);
			int deltaPixels = (int)(delta*stripWidth+1);
			//System.out.println(deltaPixels);
			//System.out.println("THUMB:"+thumb+" MAX: "+max+"SELECTION:"+selection+" THUMBEDGE:"+thumbEdge+" stripWidth:"+stripWidth+" THUMBLOC: "+thumbLoc);
			Point thumbPoint = new Point(bounds.x+BUTTON_SIZE+thumbLocPixels+thumbWidthPixels/2,bounds.y+bounds.height/2);
			Point offset = new Point(0,0);
			
			Point moveTo = new Point(bounds.x+BUTTON_SIZE+setEdgePixels+thumbWidthPixels/2,bounds.y+bounds.height/2);
			mouseMove(thumbPoint.x,thumbPoint.y);			
//			mousePress(SWT.BUTTON1);
//			mouseMove(moveTo.x,moveTo.y);
//			mouseRelease(SWT.BUTTON1);
			actionWaitForIdle(bar.getDisplay());
			
			int oldSelection = setTo;
			
			if(setTo<selection){
				while(true){
					
					Robot.syncExec(bar.getDisplay(),this,new Runnable(){
						public void run(){
							selection = bar.getSelection();
						}
					});
					
					if(oldSelection==selection){
						//System.out.println("Exiting b/c oldSelection==selection");
						break;
					}
																			
					if(selection<=setTo){
						//System.out.println("Exiting b/c selection<=setTo");
						break;
					}	
					offset.x-=deltaPixels;
					mousePress(SWT.BUTTON1);
					mouseMove(thumbPoint.x+offset.x,thumbPoint.y+offset.y);
					mouseRelease(SWT.BUTTON1);
					actionWaitForIdle(bar.getDisplay());
				
					oldSelection = selection;
				}
			}
			
			else if(setTo>selection){
				while(true){
					Robot.syncExec(bar.getDisplay(),this,new Runnable(){
						public void run(){
							selection = bar.getSelection();
						}
					});
					
					if(oldSelection==selection)
						break;
					if(selection>=setTo)
						break;
					
					offset.x+=deltaPixels;
					mousePress(SWT.BUTTON1);
					mouseMove(thumbPoint.x+offset.x,thumbPoint.y+offset.y);
					mouseRelease(SWT.BUTTON1);
					actionWaitForIdle(bar.getDisplay());
				
					oldSelection = selection;
				}
			}			
		}

		else{// style==SWT.VERTICAL
			double thumbEdge = (double)((double)selection/(double)(maximum-minimum));
			double thumbWidth = (double)((double)thumb/(double)(maximum-minimum));
			double delta = (double)((double)increment/(double)(maximum-minimum));
			int stripWidth = bounds.height - 2*BUTTON_SIZE;
			int thumbLocPixels = (int)(thumbEdge*stripWidth);
			int thumbWidthPixels = (int)(thumbWidth*stripWidth);
			int deltaPixels = (int)(delta*stripWidth+1);
			//System.out.println("THUMB:"+thumb+" MAX: "+max+"SELECTION:"+selection+" THUMBEDGE:"+thumbEdge+" stripWidth:"+stripWidth+" THUMBLOC: "+thumbLoc);
			Point thumbPoint = new Point(bounds.x+bounds.width/2,bounds.y+BUTTON_SIZE+thumbLocPixels+thumbWidthPixels/2);
			Point offset = new Point(0,0);
			mouseMove(thumbPoint.x,thumbPoint.y);			
			actionWaitForIdle(bar.getDisplay());
		
			int oldSelection = setTo;
		
			if(setTo<selection){
				while(true){
				
					Robot.syncExec(bar.getDisplay(),this,new Runnable(){
						public void run(){
							selection = bar.getSelection();
						}
					});
				
					if(oldSelection==selection){
						//System.out.println("Exiting b/c oldSelection==selection");
						break;
					}
																		
					if(selection<=setTo){
						//System.out.println("Exiting b/c selection<=setTo");
						break;
					}	
					offset.y-=deltaPixels;
					mousePress(SWT.BUTTON1);
					mouseMove(thumbPoint.x+offset.x,thumbPoint.y+offset.y);
					mouseRelease(SWT.BUTTON1);
					actionWaitForIdle(bar.getDisplay());
			
					oldSelection = selection;
				}
			}
		
			else if(setTo>selection){
				while(true){
					Robot.syncExec(bar.getDisplay(),this,new Runnable(){
						public void run(){
							selection = bar.getSelection();
						}
					});
					//System.out.println((thumbPoint.x+offset.x)+","+(thumbPoint.y+offset.y));				
					//System.out.println("setTo="+setTo+" selection="+selection);
					if(oldSelection==selection){
						//System.out.println("Exiting b/c oldSelection==selection");
						break;
					}
					if(selection>=setTo){
						//System.out.println("Exiting b/c selection>=setTo");
						break;
					}
					
					offset.y+=deltaPixels;
					mousePress(SWT.BUTTON1);
					//actionDelay(500);
					mouseMove(thumbPoint.x+offset.x,thumbPoint.y+offset.y);
					mouseRelease(SWT.BUTTON1);
					//actionDelay(500);
					actionWaitForIdle(bar.getDisplay());
			
					oldSelection = selection;
				}
			}				
		}
		
		// make sure that we actually changed the scroll position
		Robot.syncExec(bar.getDisplay(),this,new Runnable(){
			public void run(){
				selection = bar.getSelection();
			}
		});
		
//		int rangeMin = setTo - increment;
//		int rangeMax = setTo + increment;
//		if( !(selection>=rangeMin && selection<=rangeMax) )
//			Log.warn("Failed to set scrollbar appropriately (setTo="+setTo+" selection="+selection);
		
		actionWaitForIdle(bar.getDisplay());	
	}	

	/** Returns the bounds of the given ScrollBar in global display coords **/
	protected synchronized Rectangle getBounds(final ScrollBar bar){
		Robot.syncExec(bar.getDisplay(),this,new Runnable(){
			public void run(){
				parent = bar.getParent();			
				style = bar.getStyle();
			}			
		});	
			
		RECT rect = new RECT ();
		OS.GetClientRect (parent.handle, rect);
		int x = 0, y = 0, width, height;
		if ((style & SWT.HORIZONTAL) != 0) {
			y = rect.bottom - rect.top;
			width = rect.right - rect.left;
			height = OS.GetSystemMetrics (OS.SM_CYHSCROLL);
		} else {
			x = rect.right - rect.left;
			width = OS.GetSystemMetrics (OS.SM_CXVSCROLL);
			height = rect.bottom - rect.top;
		}
		
		Point topLeft = bar.getParent().toDisplay(x,y);
		Point botRight = bar.getParent().toDisplay(x+width,y+height);
		return new Rectangle (topLeft.x, topLeft.y, botRight.x-topLeft.x,botRight.y-topLeft.y);
	}
}
