package com.ibm.abbot.tester.swt;

import org.eclipse.swt.widgets.*; 
import org.eclipse.swt.graphics.*;

/**
 * @author Kevin T Dale
 * 
 * Provides a means to find the coordinates of an SWT Widget in display-space, 
 * given an SWT Display object. 
 * 
 */
public class WidgetLocator{
	
	/**
	 * Returns the location of the top left-hand corner of a the client area
	 * of a given Widget in display space.
	 * 
	 * @param w the Widget to be found
	 * @return the location of Widget w
	 */
	
	public static Point getLocation(Widget w){
		return getLocation(w,true);	
	}
	
	/**
	 * Returns the location of the top left-hand corner of a given Widget in display 
	 * space, optionally ignoring the Widget's trimmings.  For example, to obtain  
	 * the location of the area of the Widget that can recieve mouse actions, use
	 * <code>ignoreBorder==true</code>. 
	 * 
	 * @param w the Widget to be found
	 * @param ignoreBorder should the border be considered?	 
	 * @return the location of Widget w
	 */	
	public static Point getLocation(Widget w, boolean ignoreBorder){
		Rectangle bounds; 
		try{
			bounds = getBounds(w,ignoreBorder);
		}
		catch(NullPointerException npe){
			return null;
		}
		return new Point(bounds.x,bounds.y);
	}
	
	/**
	 * Finds the display-space rectangle that bounds the given Widget's client area.
	 * 
	 * @param w the Widget to be found
	 * @return the bounds of the widget's client area
	 */
	public static Rectangle getBounds(Widget w){
		return getBounds(w,true);	
	}
	/**
	 * Finds the display space rectangle that bounds the given Widget, optionally
	 * ignoring the Widget's trimmings.  
	 * 
	 * @param w the Widget to be found
	 * @param ignoreBorder should the border be considered?
	 * @return the bounding rectangle
	 */	
	public static Rectangle getBounds(Widget w, boolean ignoreBorder){
		if(w instanceof Control){
			if(ignoreBorder && w instanceof Decorations){
				Rectangle bounds = ((Decorations)w).getBounds();
				Rectangle clientArea = ((Decorations)w).getClientArea();
				Rectangle calced = ((Decorations)w).computeTrim(bounds.x,bounds.y,
												  clientArea.width,clientArea.height);
				Rectangle correct = new Rectangle(2*bounds.x-calced.x,
												  2*bounds.y-calced.y,
												  clientArea.width-1,	//bug workaround 
												  clientArea.height-1);	
				return toDisplay(correct,((Decorations)w));							
			}
			if(ignoreBorder && w instanceof Button){
				Rectangle bounds = ((Button)w).getBounds();
				bounds.width-=1;
				bounds.height-=1;
				return toDisplay(bounds,(Control)w);				
			}
			/* if implemented for any other Widget-types, you may need to do the same bug-workaround as
			 * above, where you subtract 1 from both the width and height
			 */
			else{
				return toDisplay(((Control)w).getBounds(),((Control)w));
			}
		}
		System.err.println("Unable to find coordinates of Widgets of type "+
							w.getClass()+"; returning NULL");
		return null;
	}
			
	/**
	 * Helper method to convert a Rectangle, given in the coordinate system of
	 * the given Control's parent, to display-cordinates.
	 */
	protected static Rectangle toDisplay(Rectangle r, Control c){
		Point topLeft = new Point(r.x,r.y);
		Point bottomRight = new Point(r.x+r.width,r.y+r.height);
		
		if(c.getParent()!=null){// if not a top-level shell
			topLeft = c.getParent().toDisplay(topLeft.x,topLeft.y);
			bottomRight = c.getParent().toDisplay(bottomRight.x,bottomRight.y);
		}
		
		return new Rectangle(topLeft.x,topLeft.y,bottomRight.x-topLeft.x,bottomRight.y-topLeft.y);
	}
}
