 
 /* ISSUES TO RESOLVE:
  * 
  * 1) Where should swt extensions go (as is, in abbot.tester.extensions)?
  * 	[for now, we'll leave it in abbot.tester.extensions]
  * 2) Are we to put waitForIdle() calls here, or not?
  * 	[for now, we'll include them only where they were included before]
  * 3) What params should keyPress,keyRelease,etc take?
  * 	[for now, we'll NOT use strings as with ComponentTester] 
  */

package com.ibm.abbot.tester.swt;

import java.lang.reflect.Method;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;
import java.util.HashSet;

import javax.accessibility.AccessibleContext;
import javax.accessibility.AccessibleIcon;

import abbot.BugReport;
import abbot.Condition;
import abbot.Log;
import abbot.WaitTimedOutError;

import com.ibm.abbot.*;
import com.ibm.abbot.script.*;

import org.eclipse.swt.*;
import org.eclipse.swt.widgets.*;
import org.eclipse.swt.graphics.*;
/**
 * This is the base class for tester objects.  WidgetTester primarily contains
 * three types of methods:
 * 1) action* methods, for executing a particular action on a widget
 * 2) get* methods, for obtaining info about the current state of a widget from 
 *    any thread;
 * 3) assert* methods, for making assertions about a widget's state from any thread
 */
public class WidgetTester extends Robot {	
	/* Widget properties that are obtainable by member getter functions */
	protected Object data;
	protected int style;
	protected boolean disposed;
	/**
	 * Sets the above properties to their current values for the given widget. 
	 * NOTE: This should be called in a block of code synchronized on this
	 * tester if the results of the call are to be used.
	 */
	protected synchronized void getProperties(final Widget w){
		Robot.syncExec(w.getDisplay(),this,new Runnable(){
			public void run(){
				data = w.getData();
				style = w.getStyle();
				disposed = w.isDisposed();
			}			
		});
	}
	
	/**
	 * These getter methods return a particular property of the given widget.
	 * @see the corresponding member function in class Widget   
	 */ 
	/* Begin getters */
	public synchronized Object getData(Widget w){
		getProperties(w);
		return data; 
	}	
	public synchronized int getStyle(Widget w){
		getProperties(w);
		return style;
	}
	public synchronized Object getData(final Widget w, final String key){
		Robot.syncExec(w.getDisplay(),this,new Runnable(){
			public void run(){
				data = w.getData(key);
			}
		});
		return data;
	}
	public synchronized boolean isDisposed(final Widget w){
		getProperties(w);
		return disposed;
	}
	/* End getters */	
	
	/** 
	 * Get the location of the widget in global screen coordinates 
	 **/
	public Point getGlobalLocation(Widget w){
		return getGlobalLocation(w,true);
	}
	
	Point p;
	/** 
	 * Get the location of the widget in global screen coordinates,
	 * optionally ignoring the 'trimmings'.
	 */
	public synchronized Point getGlobalLocation(final Widget w, final boolean ignoreBorder){
		Robot.syncExec(w.getDisplay(),this,new Runnable(){
			public void run(){
				p = WidgetLocator.getLocation(w,ignoreBorder);
			}
		});
		return p;
	}
	
	/** 
	 * Get the bounding rectangle for the given Widget in global
	 * screen coordinates.
	 */
	public Rectangle getGlobalBounds(Widget w){
		return getGlobalBounds(w,true);
	}
	
	Rectangle bounds;
	/** 
	 * Get the bounding rectangle for the given Widget in global
	 * screen coordinates, optionally ignoring the 'trimmings'.
	 */
	public synchronized Rectangle getGlobalBounds(final Widget w, final boolean ignoreBorder){
		Robot.syncExec(w.getDisplay(),this,new Runnable(){
			public void run(){
				bounds = WidgetLocator.getBounds(w,ignoreBorder);
			}
		});
		return bounds;
	}
	
	/** Maps class names to their corresponding Tester object. */
   private static HashMap testers = new HashMap();

   /** Establish the given WidgetTester as the one to use for the given
	* class.  This may be used to override the default tester for a given
	* core class.  Note that this will only work with widgets loaded by		////
	* the framework class loader, not those loaded by the class loader for	////
	* the code under test.													////
	*/
   public static void setTester(Class forClass, WidgetTester tester) {
	   testers.put(forClass.getName(), tester);
   }

   /** Return the appropriate Tester for the given object. */
   public static WidgetTester getTester(Widget widget) {
	   return widget != null ? getTester(widget.getClass())
		   : getTester(Widget.class);
   }

   /** Find the corresponding Tester object for the given widget class,
	   chaining up the inheritance tree if no specific tester is found for
	   that class.<p>
	   The abbot tester package is searched first, followed by the tester
	   extensions package.
   */
   public static WidgetTester getTester(Class widgetClass) {
	   Log.debug("Looking up tester for " + widgetClass);
	   if (!Widget.class.isAssignableFrom(widgetClass)) {
		   String msg = "Class " + widgetClass.getName()
			   + " is not derived from org.eclipse.swt.widgets.Widget";
		   throw new IllegalArgumentException(msg);
	   }
	   WidgetTester tester = (WidgetTester)
		   testers.get(widgetClass.getName());
	   if (tester == null) {
		   String cname = simpleClassName(widgetClass) + "Tester";
		   String pkg = WidgetTester.class.getPackage().getName();
		   tester = findTester(pkg + "." + cname, widgetClass);
		   if (tester == null) {
			   tester = findTester(pkg + ".extensions." + cname,
			widgetClass);
			   if (tester == null) {
				   tester = getTester(widgetClass.getSuperclass());
			   }
		   }
		   if (tester != null && !tester.isExtension()) {
			   // Only cache it if it's part of the standard framework,
			   // but cache it for every level that we looked up, so we
			   // don't repeat the effort.
			   testers.put(widgetClass.getName(), tester);
		   }
	   }

	   return tester;
   }

   //\/(extensions to swt stuff may go elsewhere)
   /** Return whether this tester is an extension. */
   public final boolean isExtension() {
	   return getClass().getName().startsWith("abbot.tester.extensions");
   }   
     
   /** Look up the given class, using special class loading rules to maintain
	  framework consistency. */
 	private static Class resolveClass(String testerName, Class widgetClass)
	  	throws ClassNotFoundException {
	  	// Extension testers must be loaded in the context of the code under
	  	// test.
	  	Class cls;
	  	if (testerName.startsWith("abbot.tester.extensions")) {
		  	cls = Class.forName(testerName, true,
								widgetClass.getClassLoader());
	  	}
	  	else {
		  	cls = Class.forName(testerName);
	  	}
	  	Log.debug("Loaded class " + testerName + " with "
				  + cls.getClassLoader());
	  	return cls;
  	}
	///\
	
	/** Look up the given class with a specific class loader. */
	private static WidgetTester findTester(String testerName,
											  Class widgetClass) {
		WidgetTester tester = null;
		Class testerClass = null;
		try {
			testerClass = resolveClass(testerName, widgetClass);
			tester = (WidgetTester)testerClass.newInstance();
		}
		catch(InstantiationException ie) {
			Log.warn(ie);
		}
		catch(IllegalAccessException iae) {
			Log.warn(iae);
		}
		catch(ClassNotFoundException cnf) {
			//Log.debug("Class " + testerName + " not found");
		}
		catch(ClassCastException cce) {
			throw new BugReport("Class loader conflict: environment "
								+ WidgetTester.class.getClassLoader() 
								+ " vs. " + testerClass.getClassLoader());
		}

		return tester;
	}
	
	protected WidgetFinder getFinder() {
		return DefaultWidgetFinder.getFinder();
	}

	/** Derive a tag from the given accessible context if possible, or return
	* null.
	*/
   protected String deriveAccessibleTag(AccessibleContext context) {
	   String tag = null;
	   if (context != null) {
		   if (context.getAccessibleName() != null) {
			   tag = context.getAccessibleName();
		   }
		   if ((tag == null || "".equals(tag))
			   && context.getAccessibleIcon() != null
			   && context.getAccessibleIcon().length > 0) {
			   AccessibleIcon[] icons = context.getAccessibleIcon();    
			   tag = icons[0].getAccessibleIconDescription();
			   if (tag != null) {
				   tag = tag.substring(tag.lastIndexOf("/") + 1);
				   tag = tag.substring(tag.lastIndexOf("\\") + 1);
			   }
		   }
	   }
	   return tag;
   }

  	private static final String[] tagMethods = {
	  	"getText",
	  	"getData" 
  	};

	private static final Class[][] tagParamTypes = {
		null,
		{String.class}
	};
	
	private static final Object[][] tagArgs = {
		null,
		{"name"}  	
	};
	
  	/** Return a reasonable identifier for the given widget. */
  	public static String getTag(Widget widget) {
		return getTester(widget.getClass()).deriveTag(widget);
  	}

	/** Provide a String that is fairly distinct for the given widget.  For
	 * a generic widget, attempt to look up some common patterns such as a
	 * title or label.  Derived classes should absolutely override this method
	 * if such a String exists.<p>
	 * Don't use widget names as tags.<p>
	 */
	public String deriveTag(Widget widget) {
		Method m = null;
		String tag = null;
		// Try a few default methods
		for (int i=0;i < tagMethods.length;i++) {
			// Don't use getText on text components
			if ((widget instanceof Text)
				&& "getText".equals(tagMethods[i])) {
				continue;
			}
			try {
				m = widget.getClass().getMethod(tagMethods[i], tagParamTypes[i]);
				String tmp = (String)m.invoke(widget, tagArgs[i]);
				// Don't ever use empty strings for tags
				if (tmp != null && !"".equals(tmp)) {
					tag = tmp;
					break;
				}
			}
			catch(Exception e) {
				// System.err.println("tagMethods["+i+"] = "+m);
				// System.err.println(tagMethods[i]);
				// System.err.println(tagParamTypes[i]);
				// System.err.println(tagArgs[i]);
			}
		}
		// In the absence of any other tag, try to derive one from something
		// recognizable on one of its ancestors.
		if (tag == null || "".equals(tag)) {
			Widget parent = DefaultWidgetFinder.getFinder().getWidgetParent(widget);
			if (parent != null) {
				String ptag = getTag(parent);
				if (ptag != null && !"".equals(tag)) {
					// Don't use the tag if it's simply the window title; that
					// doesn't provide any extra information.
					if (!ptag.endsWith(" Root Pane")) {
						StringBuffer buf = new StringBuffer(ptag);
						int under = ptag.indexOf(" under ");
						if (under != -1)
							buf = buf.delete(0, under + 7);
						buf.insert(0, " under ");
						buf.insert(0, simpleClassName(widget.getClass()));
						tag = buf.toString();
					}
				}
			}
		}

		return tag;
	}
	

	/**
	 * Wait for an idle AWT event queue.  Will return when there are no more
	 * events on the event queue.  
	 */
	public void actionWaitForIdle() {
		waitForIdle();
	}

	/**
	 * Wait for an idle AWT event queue.  Will return when there are no more
	 * events on the event queue for the given display. 
	 */
	public void actionWaitForIdle(Display d){
		waitForIdle(d);	
	}

	/** Delay the given number of ms. */
	public void actionDelay(int ms) {
		delay(ms);
	}

	public void actionSelectMenuItemByText(Menu menu, String text) {
		selectMenuItemByText(menu, text);
	}
	
	// FIXME do actionSelectMenuItemByName in addition to by component
	// FIXME do actionSelectMenuItemByPath (for dynamics) which follows labels
	/** Select the given menu item. */
	public void actionSelectMenuItem(MenuItem item) {
		Log.debug("Attempting to select menu item " + toString(item));
		selectMenuItem(item);
	}

	/** Open the item's popup menu at the given coordinates of its parent control, and
	 *  select the given item.
	 */
	public void actionSelectPopupMenuItem(MenuItem item, int x, int y){
		selectPopupMenuItem(item,x,y);
	}

	/** Click on the center of the widget. */
	public void actionClick(Widget widget) {
		click(widget);
		waitForIdle(widget.getDisplay());
	}

	/** Click on the widget at the given location. */
	public void actionClick(Widget widget, int x, int y) {
		click(widget, x, y, SWT.BUTTON1);
		waitForIdle(widget.getDisplay());
	}

	/** Click on the widget at the given location.  The buttons string
	 * should be the org.eclipse.swt.SWT field name for the mask.
	 */
	public void actionClick(Widget widget, int x, int y, String buttons) {
		click(widget,x,y,getModifiers(buttons));
		waitForIdle(widget.getDisplay());
	}

	/** 
	 * Click on the widget at the given location.  The buttons string
	 * should be the org.eclipse.swt.SWT field name for the mask.  This 
	 * variation provides for multiple clicks. 
	 */
	public void actionClick(Widget comp, int x, int y,
							String buttons, int count) {
		click(comp, x, y, getModifiers(buttons), count);
		waitForIdle(comp.getDisplay());
	}
	
	/*TODO implement Robot.getKeyCode(String) and change these keyMethods to take
	 * Strings as parameters
	 */
	
	/**
	 * Press the keys contained in the given accelerator
	 */
	public void actionKeyPress(int accelerator,Display d){
		keyPress(accelerator);
		
		waitForIdle(d);	
	}

	/**
	 * Release the keys contained in the given accelerator
	 */
	public void actionKeyRelease(int accelerator,Display d) {
		keyRelease(accelerator);
		waitForIdle(d);
	}

	/**
	 * Press, and release, the keys contained in the given accelerator
	 */
	public void actionKey(int accelerator,Display d){
		key(accelerator);
		waitForIdle(d);
	}

	/**
	 * Type the given character.  Note that this sends the key to whatever
	 * component currently has the focus.
	 */	
	public void actionKeyChar(char c,Display d){
		key((int)c);
		waitForIdle(d);	
	}
	
	/** 
	 * Type the given string. 
	 */
	public void actionKeyString(String string,Display d){
		keyString(string);
		waitForIdle(d);
	}
	

	/** Set the focus on to the given component. */
	/* TODO MAY NEED TO CHECK THAT THE CONTROL DOES INDEED HAVE FOCUS */
	public void actionFocus(Widget widget) {
		while(!(widget instanceof Control))
			widget = DefaultWidgetFinder.getFinder().getWidgetParent(widget);
		focus((Control)widget);
		waitForIdle(widget.getDisplay());
	}

	/** Perform a drag action.  Derived classes should provide more specific
	 * identifiers for what is being dragged, e.g. actionDragTableCell or
	 * actionDragListElement.
	 */	
	public void actionDrag(Widget source, int x, int y){
		drag(source,x,y,SWT.BUTTON1);
		waitForIdle(source.getDisplay());
	}

	/** Perform a drag action.  Derived classes should provide more specific
	 * identifiers for what is being dragged, e.g. actionDragTableCell or
	 * actionDragListElement.   The modifiers represents the set of active
	 * modifiers when the drop is made.
	 */
	public void actionDrag(Widget source, int x, int y, int modifiers){
		drag(source,x,y,modifiers);
		waitForIdle(source.getDisplay());
	}

	/** Perform a basic drop action (implicitly causing a preceding mouse
	 * drag).
	 */ 	
	public void actionDrop(Widget target, int x, int y){
		drop(target,x,y,SWT.BUTTON1);
		waitForIdle(target.getDisplay());
	}
	
	/** Perform a basic drop action (implicitly causing a preceding mouse
	 * drag).  The modifiers represents the set of active modifiers when the
	 * drop is made.
	 */ 
	public void actionDrop(Widget target, int x, int y, int modifiers){
		drop(target,x,y,modifiers);
		waitForIdle(target.getDisplay());
	}
	
	/** Return whether the widget's contents matches the given image. */
	public boolean assertImage(Widget widget, Image image,
									  boolean ignoreBorder) {
		Image widgetImage = capture(widget, ignoreBorder);
		SWTImageComparator ic = new SWTImageComparator();
		boolean result = ic.compare(image, widgetImage) == 0;
		widgetImage.dispose();
		return result;
	}

	
	/** Wait for the given condition, throwing an ActionFailedException if it
	 * times out.
	 */
	protected void waitAction(String desc, Condition cond)
		throws ActionFailedException {
		try { wait(cond); }
		catch(WaitTimedOutError wto) {
			throw new ActionFailedException(desc);
		}
	}	
	
	/** Returns whether Decorations corresponding to the given String is
	 * showing.  The string may be a plain String or regular expression and
	 * may match either the Decoration's title or name
	 */ 
	// FIXME provide more options for identifying the window (name, title,
	// class, nth window)		
	public synchronized boolean assertDecorationsShowing(String title) {
		final Decorations w = getFinder().findDecorations(title);
		boolT = false;
		if (w==null)
			return false;
		Robot.syncExec(w.getDisplay(),this,new Runnable(){
			public void run(){
				boolT = w.isVisible();
			}
		});
	
		return  boolT;
	 }

	/** Convenience wait for a window to be displayed.  The given string may
	  * be a plain String or regular expression and may match either the window
	  * title or its Widget name.  This method is provided as a convenience
	  * for hand-coded tests, since scripts will use a wait step instead.  The
	  * property abbot.robot.component_delay affects the default timeout
	  * (default is 30s).  
	  */
	 public static void waitForFrameShowing(final String title) {
		 wait(new Condition() {
			 public boolean test() {
				 return getTester(Widget.class).assertDecorationsShowing(title);
			 }
			 public String toString() { return title + " to show"; }
		 }, componentDelay);
	 }	
	 
	/** Return whether the Widget represented by the given
		WidgetReference is available.
	*/
	public boolean assertWidgetShowing(WidgetReference ref) {
		try {
			getFinder().findWidget(ref);
			return true;
		}
		catch(Exception e) {
			return false;
		}
	}
	
	/** Wait for the Widget represented by the given WidgetReference to
		become available.  The timeout is affected by
		abbot.robot.component_delay, which defaults to 30s. 
	*/
	public static void waitForWidgetShowing(final WidgetReference ref) {
		wait(new Condition() {
			public boolean test() {
				return getTester(Widget.class).assertWidgetShowing(ref);
			}
			public String toString() { return ref + " to show"; }
		}, componentDelay);
	}


	private Method[] cachedMethods = null;

	/** Look up methods with the given prefix. */
	private Method[] getMethods(String prefix, Class returnType, 
								boolean onWidget) {
		if (cachedMethods == null) {
			cachedMethods = getClass().getMethods();
		}
		ArrayList methods = new ArrayList();
		HashSet names = new HashSet();
		Method[] mlist = cachedMethods;
		for (int i=0;i < mlist.length;i++) {
			String name = mlist[i].getName();
			if (!names.contains(name)) {
				Class[] params = mlist[i].getParameterTypes();
				if (name.startsWith(prefix)
					&& (returnType == null 
						|| returnType.equals(mlist[i].getReturnType()))
					&& ((params.length == 0 && !onWidget)
						|| (params.length > 0 
							&& (Widget.class.isAssignableFrom(params[0]) 
								== onWidget)))) {
					methods.add(mlist[i]);
					names.add(name);
				}
			}
		}
		return (Method[])methods.toArray(new Method[methods.size()]);
	}

	private Method[] cachedActions = null;
	/** Return a list of all actions defined by this class that don't depend
	 * on a widget argument. 
	 */
	public Method[] getActions() {
		if (cachedActions == null) {
			cachedActions = getMethods("action", void.class, false);
		}
		return cachedActions;
	}
	

	private Method[] cachedComponentActions = null;
	/** Return a list of all actions defined by this class that require
	 * a widget argument. 
	 */
	public Method[] getWidgetActions() {
		if (cachedComponentActions == null) {
			cachedComponentActions = getMethods("action", void.class, true);
		}
		return cachedComponentActions;
	}

    
	private Method[] cachedPropertyMethods = null;
	/** Return an array of all property check methods defined by this class.
	 * The first argument <b>must</b> be a Widget.
	 */
	public Method[] getPropertyMethods() {
		if (cachedPropertyMethods == null) {
			ArrayList all = new ArrayList();
			all.addAll(Arrays.asList(getMethods("is", boolean.class, true)));
			all.addAll(Arrays.asList(getMethods("get", null, true))); 
			// Remove getXXX or isXXX methods which aren't property checks
			Class[] args = new Class[] { Widget.class };
			try {
				all.remove(getClass().getMethod("getTag", args));
				all.remove(getClass().getMethod("getTester", args));
				all.remove(getClass().getMethod("isOnPopup", args));
			}
			catch(NoSuchMethodException e) {
			}
			cachedPropertyMethods =
				(Method[])all.toArray(new Method[all.size()]);
		}
		return cachedPropertyMethods;
	}

	private Method[] cachedAssertMethods = null;
	/** Return a list of all assertions defined by this class that don't
	 * depend on a widget argument.
	 */
	public Method[] getAssertMethods() {
		if (cachedAssertMethods == null) {
			cachedAssertMethods = getMethods("assert", boolean.class, false);
		}
		return cachedAssertMethods;
	}

	private Method[] cachedComponentAssertMethods = null;
	/** Return a list of all assertions defined by this class that require a
	 * widget argument.
	 */
	public Method[] getComponentAssertMethods() {
		if (cachedComponentAssertMethods == null) {
			cachedComponentAssertMethods =
				getMethods("assert", boolean.class, true);
		}
		return cachedComponentAssertMethods;
	}

	/** Quick and dirty strip raw text from html, for getting the basic text
		from html-formatted labels and buttons.  Behavior is undefined for
		badly formatted html.
	*/
	public static String stripHTML(String str) {
		if (str != null
			&& (str.startsWith("<html>")
				|| str.startsWith("<HTML>"))) {
			while (str.startsWith("<")) {
				int right = str.indexOf(">");
				if (right == -1)
					break;
				str = str.substring(right + 1);
			}
			while (str.endsWith(">")) {
				int right = str.lastIndexOf("<");
				if (right == -1)
					break;
				str = str.substring(0, right);
			}
		}
		return str;
	}
	
	/** Return the Widget class that corresponds to this WidgetTester
		class.  
	*/
	public Class getTestedClass(Class cls) {
		while (getTester(cls.getSuperclass()) == this) {
			cls = cls.getSuperclass();
		}
		return cls;
	}	
}
